--/topic Linking to external code
--/info
--Using dynamic libraries and C routines with Euphoria
--
--<p><code><span style="color: blue;">include</span> adv/dll.e</code></p>

namespace dll

include adv/machine.e
include std/dll.e as std_dll
include std/error.e
include std/io.e
include std/map.e
include std/sequence.e
include std/types.e
include std/utils.e
without warning

--/topic C Type Constants
--/parent Linking to external code
--/info
--Most common C, Win32, and POSIX types are included here.
--
--See also: /"C String Constants", /"Euphoria Type Constants"

public constant
--/topic C Type Constants
--/const C_CHAR
--/info
--An 8-bit C /b char.
	C_CHAR			= std_dll:C_CHAR,
--/topic C Type Constants
--/const C_BYTE
--/info
--An 8-bit C /b byte.
	C_BYTE			= std_dll:C_BYTE,
--/topic C Type Constants
--/const C_UCHAR
--/info
--An 8-bit C /b "unsigned char".
	C_UCHAR			= std_dll:C_UCHAR,
--/topic C Type Constants
--/const C_UBYTE
--/info
--An 8-bit C /b "unsigned byte".
	C_UBYTE			= std_dll:C_UBYTE,
--/topic C Type Constants
--/const C_SHORT
--/info
--A 16-bit C /b short.
	C_SHORT			= std_dll:C_SHORT,
--/topic C Type Constants
--/const C_WORD
--/info
--A 16-bit C /b WORD.
	C_WORD			= std_dll:C_WORD,
--/topic C Type Constants
--/const C_USHORT
--/info
--A 16-bit C /b "unsigned short".
	C_USHORT		= std_dll:C_USHORT,
--/topic C Type Constants
--/const C_INT
--/info
--A 32-bit C /b int.
	C_INT			= std_dll:C_INT,
--/topic C Type Constants
--/const C_BOOL
--/info
--A 32-bit C /b bool.
	C_BOOL			= std_dll:C_INT,
--/topic C Type Constants
--/const C_UINT
--/info
--A 32-bit C /b "unsigned int".
	C_UINT			= std_dll:C_UINT,
--/topic C Type Constants
--/const C_SIZE_T
--/info
--A 32-bit C /b size_t. Used for sizes of objects.
	C_SIZE_T		= std_dll:C_UINT,
--/topic C Type Constants
--/const C_LONG
--/info
--A 32-bit C /b long.
	C_LONG			= std_dll:C_INT,
--/topic C Type Constants
--/const C_ULONG
--/info
--A 32-bit C /b "unsigned long".
	C_ULONG			= std_dll:C_UINT,
--/topic C Type Constants
--/const C_POINTER
--/info
--Any 32-bit C /i pointer.
	C_POINTER		= std_dll:C_UINT,
--/topic C Type Constants
--/const C_HANDLE
--/info
--A 32-bit C /b HANDLE.
	C_HANDLE		= std_dll:C_UINT,
--/topic C Type Constants
--/const C_HWND
--/info
--A 32-bit C /b HWND.
	C_HWND			= std_dll:C_UINT,
--/topic C Type Constants
--/const C_DWORD
--/info
--A 32-bit C /b DWORD.
	C_DWORD			= std_dll:C_UINT,
--/topic C Type Constants
--/const C_WPARAM
--/info
--A 32-bit C /b WPARAM.
	C_WPARAM		= std_dll:C_LONG,
--/topic C Type Constants
--/const C_LPARAM
--/info
--A 32-bit C /b LPARAM.
	C_LPARAM		= std_dll:C_LONG,
--/topic C Type Constants
--/const C_HRESULT
--/info
--A 32-bit C /b HRESULT.
	C_HRESULT		= std_dll:C_LONG,
--/topic C Type Constants
--/const C_FLOAT
--/info
--A 32-bit C /b float.
	C_FLOAT			= and_bits( std_dll:C_FLOAT,  #0100 ),
--/topic C Type Constants
--/const C_DOUBLE
--/info
--A 64-bit C /b double.
	C_DOUBLE		= and_bits( std_dll:C_DOUBLE, #0200 ),
--/topic C Type Constants
--/const C_DWORDLONG
--/info
--A 64-bit C /b DWORDLONG.
	C_DWORDLONG		= std_dll:C_DWORDLONG,
$

public constant
--/topic C Type Constants
--/const C_CLOCK_T
--/info
--A 32-bit C /b clock_t. Used for system times in clock ticks or /i "clocks per sec".
	C_CLOCK_T		= std_dll:C_LONG,
--/topic C Type Constants
--/const C_DEV_T
--/info
--A 16-bit C /b udev_t. Used for device IDs.
	C_DEV_T			= std_dll:C_SHORT,
--/topic C Type Constants
--/const C_GID_T
--/info
--An 8-bit C /b gid_t. Used for group IDs.
	C_GID_T			= std_dll:C_CHAR,
--/topic C Type Constants
--/const C_INO_T
--/info
--A 32-bit C /b ino_t. Used for file serial numbers.
	C_INO_T			= std_dll:C_ULONG,
--/topic C Type Constants
--/const C_MODE_T
--/info
--A 16-bit C /b mode_t. Used for some file attributes.
	C_MODE_T		= std_dll:C_USHORT,
--/topic C Type Constants
--/const C_NLINK_T
--/info
--A 16-bit C /b nlink_t. Used for link counts.
	C_NLINK_T		= std_dll:C_SHORT,
--/topic C Type Constants
--/const C_OFF_T
--/info
--A 32-bit C /b off_t. Used for file sizes.
	C_OFF_T			= std_dll:C_ULONG,
--/topic C Type Constants
--/const C_PID_T
--/info
--A 32-bit C /b pid_t. Used for process IDs and process group IDs.
	C_PID_T			= std_dll:C_INT,
--/topic C Type Constants
--/const C_SSIZE_T
--/info
--A 32-bit /b ssize_t. Used for a count of bytes or an error indication.
	C_SSIZE_T		= std_dll:C_INT,
--/topic C Type Constants
--/const C_TIME_T
--/info
--A 32-bit /b time_t. Used for time in seconds.
	C_TIME_T		= std_dll:C_LONG,
--/topic C Type Constants
--/const C_UID_T
--/info
--A 16-bit C /b uid_t. Used for user IDs.
	C_UID_T			= std_dll:C_SHORT,
--/topic C Type Constants
--/const C_USECONDS_T
--/info
--A 32-bit C /b useconds_t. Used for time in microseconds.
	C_USECONDS_T	= std_dll:C_LONG,	
--/topic C Type Constants
--/const C_U8_T
--/info
--An 8-bit C /b u8_t.
	C_U8_T			= std_dll:C_UCHAR,
--/topic C Type Constants
--/const C_U16_T
--/info
--A 16-bit C /b u16_t.
	C_U16_T			= std_dll:C_USHORT,
--/topic C Type Constants
--/const C_U32_T
--/info
--A 32-bit C /b u32_t.
	C_U32_T			= std_dll:C_ULONG,
--/topic C Type Constants
--/const C_I8_T
--/info
--An 8-bit C /b i8_t.
	C_I8_T			= std_dll:C_CHAR,
--/topic C Type Constants
--/const C_I16_T
--/info
--A 16-bit C /b i16_t.
	C_I16_T			= std_dll:C_SHORT,
--/topic C Type Constants
--/const C_I32_T
--/info
--A 32-bit C /b i32_t.
	C_I32_T			= std_dll:C_LONG,
$

--/topic C String Constants
--/parent Linking to external code
--/info
--These types indicate string pointers. When you declare argument and return types with these values,
-- /c_func and /c_proc will accept strings as parameters and automatically allocate, free, and return
-- those values instead of just the pointers.
--
--See also: /"C Type Constants", /"Euphoria Type Constants"

public constant
--/topic C String Constants
--/const C_STRING
--/info
--A C-style null-terminated /b char string /i pointer (e.g. /b "char*").
--
--See also: /allocate_string, /peek_string
	C_STRING		= or_bits( std_dll:C_POINTER, #010000 ),
--/topic C String Constants
--/const C_WSTRING
--/info
--A C-style null-terminated /b wchar_t string /i pointer (e.g. /b "wchar_t*").
--
--See also: /allocate_wstring, /peek_wstring
	C_WSTRING		= or_bits( std_dll:C_POINTER, #020000 ),
--/topic C String Constants
--/const C_LPCSTR
--/info
--A C-style null-terminated /b char string /i pointer (e.g. /b "char*").
--
--See also: /allocate_string, /peek_string
	C_LPCSTR		= C_STRING,
--/topic C String Constants
--/const C_LPCWSTR
--/info
--A C-style null-terminated /b wchar_t string /i pointer (e.g. /b "wchar_t*").
--
--See also: /allocate_wstring, /peek_wstring
	C_LPCWSTR		= C_WSTRING,
$

--/topic Euphoria Type Constants
--/parent Linking to external code
--/info
--These constants are for passing Euphoria types to/from translated libraries.
--
--See also: /"C String Constants", /"C Type Constants"

public constant
--/topic Euphoria Type Constants
--/const E_INTEGER
--/info
--A Euphoria /b integer
	E_INTEGER		= std_dll:E_INTEGER,
--/topic Euphoria Type Constants
--/const E_ATOM
--/info
--A Euphoria /b atom
	E_ATOM			= std_dll:E_ATOM,
--/topic Euphoria Type Constants
--/const E_SEQUENCE
--/info
--A Euphoria /b sequence
	E_SEQUENCE		= std_dll:E_SEQUENCE,
--/topic Euphoria Type Constants
--/const E_OBJECT
--/info
--A Euphoria /b object
	E_OBJECT		= std_dll:E_OBJECT,
$

--/topic Linking to external code
--/const NULL
--/info
--A null pointer.
public constant NULL = std_dll:NULL

public constant
	C_MASK_SIZE		= #000000FF,
	C_MASK_FLOAT	= #0000FF00,
	C_MASK_STRING	= #00FF0000,
	C_MASK_TYPE		= #FF000000,
$

map m_libs	= map:new()	-- library filenames
map m_args	= map:new()	-- func/proc argument types
map m_retv	= map:new()	-- func/proc return value types
map m_vars	= map:new()	-- variable names
map m_cbid	= map:new() -- callback routine ids

ifdef WINDOWS then
map m_cdecl	= map:new()	-- cdecl flags
end ifdef

--/topic Linking to external code
--/func open_dll( sequence file_name, types:boolean use_cdecl = 0 )
--/info
--Open a Windows dynamic link library (.dll) file, or a Unix shared library (.so) file.
--
--<dl>
--<dt> /b Parameters </dt>
--
--<ol>
--<li> /b file_name : a /b sequence, the name of the shared library to open or a sequence of filenames
-- to try and open.</li>
--<li> /b use_cdecl : an /b integer, if non-zero, then force all routines in this library to be declared
-- as CDECL instead of STDCALL</li>
--</ol>
--
--<dt> /b Returns </dt>
--
--<dd>An /b atom, actually a 32-bit address. 0 is returned if the library cannot be loaded.</dd>
--
--<dt> /b Comments </dt>
--
--<dd>The /b use_cdecl flag only applies to Windows. Setting it on other platforms has no effect.</dd>
--
--<dd>By default, /open_dll will crash your application if it cannot load a library. See
-- /"Errors and debugging" for more details.</dd>
--</dl>
--
--See also: /define_c_var, /define_c_func, /define_c_proc, /c_func, /c_proc
public function open_dll( sequence file_name, types:boolean use_cdecl = 0 )
	
	-- ensure we have an array of strings
	-- so we can loop through the values
	if string( file_name ) then
		file_name = {file_name}
	end if
	
	atom lib = 0
	integer found = 0
	
	for i = 1 to length( file_name ) do
		
		-- attempt to link the library
		lib = std_dll:open_dll( file_name[i] )
		
		if lib != 0 then
			
			-- store the index and exit
			found = i
			exit
			
		end if
		
	end for
	
	if found = 0 then
		-- the library was not found
		
		-- build the warning message
		sequence message = sprintf( "open_dll: unable to link %s\n", {stdseq:join(file_name, " or ")} )
		
ifdef DLL_WARNING then
		-- just issue a warning to the console
		puts( STDERR, message )
elsifdef not DLL_SILENT then
		-- crash altogether
		crash( message )
end ifdef
		
	else
		-- the library was found
		
		-- store a reverse lookup
		map:put( m_libs, lib, file_name[found] )
		
ifdef WINDOWS then
ifdef DLL_CDECL then
		-- always use cdecl
		use_cdecl = 1
end ifdef
		-- store the cdecl flag
		map:put( m_cdecl, lib, use_cdecl )
end ifdef
		
ifdef DLL_DEBUG then
		
		-- build the debug message
		sequence message = "open_dll: %s = #%x"
		
ifdef WINDOWS then
		if use_cdecl then
			-- indicate we are using cdecl
			message &= " (using cdecl)"
		end if
end ifdef
		
		message &= "\n"
		
		-- print the debug message to the console
		printf( STDERR, message, {file_name[found],lib} )
		
end ifdef
		
	end if
	
	return lib
end function

--/topic Linking to external code
--/func define_c_var( atom lib, sequence variable_name )
--/info
--Gets the memory address of a symbol in a shared library.
--
--<dl>
--<dt> /b Parameters </dt>
--
--<ol>
--<li> /b lib : an /b atom, the address of the shared library from /open_dll
--<li> /b variable_name : a /b sequence, the name of the exported C variable in the library
--</ol>
--
--<dt> /b Returns </dt>
--
--<dd>An /b atom, the memory address of /b variable_name.</dd>
--
--<dt> /b Comments </dt>
--
--<dd>Once you have the address of a C variable, and you know its type, you can use /peek and /poke
-- to read or write the value of the variable.
--
--<dd>By default, /define_c_var will crash your application if it cannot find the variable. See
-- /"Errors and debugging" for more details.</dd>
--</dl>
--
--See also: /peek, /peeks, /peek2s, /peek2u, /peek4s, /peek4u, /peek_string, /peek_wstring, /poke,
-- /poke2, /poke4, /poke_string, /poke_wstring
public function define_c_var( atom lib, sequence variable_name )
	
	-- get the variable address
	atom addr = std_dll:define_c_var( lib, variable_name )
	
	if addr = 0 then
		-- the variable was not found
		
		-- get the library name
		sequence file_name = map:get( m_libs, lib, "unknown" )
		
		-- build the warning message
		sequence message = sprintf( "define_c_var: unable to find %s in %s\n", {variable_name,file_name} )
		
ifdef DLL_WARNING then
		-- just issue a warning to the console
		puts( STDERR, message )
elsifdef not DLL_SILENT then
		-- crash altogether
		crash( message )
end ifdef
	
	else
		-- the variable was found
		
ifdef DLL_DEBUG then
		-- print the debug message to the console
		printf( STDERR, "define_c_var: %s = #%x\n", {variable_name,addr} )
end ifdef
		
	end if
	
	return addr
end function

--/topic Linking to external code
--/func define_c_func( object lib, object routine_name, sequence arg_types, atom return_type )
--/info
--Define the paramaters of a C function or machine-code routine that returns a value.
--
--<dl>
--<dt> /b Parameters </dt>
--
--<ol>
--<li> /b lib : an /b object, either an entry point returned by /open_dll, or "" to denote a routine
-- in memory.</li>
--<li> /b routine_name : an /b object, either the name of a function in the shared library or the
-- memory address of the machine-code routine.</li>
--<li> /b arg_types : a /b sequence, a list of /"C Type constants" for each argument.</li>
--<li> /b return_type : an /b atom, the type constant or the return value.</li>
--</ol>
--
--<dt> /b Returns </dt>
--
--<dd>A small /b integer, the routine id of the function.</dd>
--
--<dt> /b Comments </dt>
--
--<dd>The length of /b routine_name should not exceed 1,024 characters.</dd>
--
--<dd>Use the returned value as the first parameter to /c_func or /c_proc.
-- A returned value of -1 indicates that the function could not be found.</dd>
--
--<dd>On Windows, you can add a '+' character as a prefix to the function name. This indicates to
-- Euphoria that the function uses the CDECL calling convention. By default, Euphoria assumes that C
-- routines accept STDCALL convention on Windows. See /open_dll on how to force this for all
-- functions in a given library.</dd>
--
--<dd>If you are not interested in using the value returned by the C function, you should instead
-- define it with /define_c_proc and call it with /c_proc.</dd>
--
--<dd>By default, /define_c_func will crash your application if it cannot find the function. See
-- /"Errors and debugging" for more details.</dd>
--</dl>
--
--See also: /define_c_proc, /c_func, /c_proc
public function define_c_func( object lib, object routine_name, sequence arg_types, atom return_type )
	
	-- we use this routine for funcs and procs
	sequence function_name = iff( return_type, "define_c_func", "define_c_proc" )
	
	-- store the original name
	object original_name = routine_name
	if not sequence( original_name ) then
		-- this is actually a memory address, so format it
		original_name = sprintf( "#%08x", original_name )
	end if
	
ifdef WINDOWS then
	
ifdef DLL_CDECL then
	-- always use cdecl
	integer use_cdecl = 1
elsedef
	-- get the cdecl flag for this library (default = 0)
	integer use_cdecl = map:get( m_cdecl, lib, 0 )
end ifdef
	
	-- add '+' to the routine name for cdecl
	if use_cdecl then
		routine_name = '+' & routine_name
	end if
	
end ifdef
	
	-- store the original values
	sequence original_args = arg_types
	atom original_retv = return_type
	
	for i = 1 to length( arg_types ) do
		
		-- remove our indicators from the argument types
		
		if and_bits( arg_types[i], C_MASK_STRING ) then
			arg_types[i] = and_bits( arg_types[i], not_bits(C_MASK_STRING) )
		end if
		
		if and_bits( arg_types[i], C_MASK_FLOAT ) then
			arg_types[i] = and_bits( arg_types[i], not_bits(C_MASK_FLOAT) )
		end if
		
	end for
	
	-- remove our 'string' indicator from the return type
	
	if and_bits( return_type, C_MASK_STRING ) then
		return_type = and_bits( return_type, not_bits(C_MASK_STRING) )
	end if
	
	if and_bits( return_type, C_MASK_FLOAT ) then
		return_type = and_bits( return_type, not_bits(C_MASK_FLOAT) )
	end if
	
	-- link the function
	integer id = std_dll:define_c_func( lib, routine_name, arg_types, return_type )
	
	if id = -1 then
		-- the function was not found
		
		-- get the library name
		sequence file_name = map:get( m_libs, lib, "unknown" )
		
		-- build the warning message
		sequence message = sprintf( "%s: unable to link %s in %s\n", {function_name,original_name,file_name} )
		
ifdef DLL_WARNING then
		-- just issue a warning to the console
		puts( STDERR, message )
elsifdef not DLL_SILENT then
		-- crash altogether
		crash( message )
end ifdef
		
	else
		-- the function was found
		
ifdef DLL_DEBUG then
		-- build the debug message
		sequence message = sprintf( "%s: %%s = %%d\n", {function_name} )
		
		-- print the debug message to the console
		printf( STDERR, message, {original_name,id} )
end ifdef
		
		-- store the argument types and return type
		map:put( m_args, id, original_args )
		map:put( m_retv, id, original_retv )
		
	end if
	
	return id
end function

--/topic Linking to external code
--/func define_c_proc( object lib, object routine_name, sequence arg_types )
--/info
--Define the paramaters of a C function or machine-code routine does not return a value.
--
--See also: /define_c_func, /c_func, /c_proc
public function define_c_proc( object lib, object routine_name, sequence arg_types )
	-- just call define_c_func with no return type
	return define_c_func( lib, routine_name, arg_types, 0 )
end function

--/topic Linking to external code
--/func c_func( integer rid, sequence args = {} )
--/info
--Call a C function, machine code routine, or translated/compiled Euphoria routine by routine id.
--
--<dl>
--<dt> /b Parameters </dt>
--
--<ol>
--<li> /b rid : an /b integer, a routine id from /define_c_func or /define_c_proc.</li>
--<li> /b args : a /b sequence, the list of parameters to pass to the function.</li>
--</ol>
--
--<dt> /b Returns </dt>
--
--<dd>An /b object, whose type and meaning was defined by calling /define_c_func or /define_c_proc.</dd>
--
--<dt> /b Comments </dt>
--
--<dd>If you declare the arguments of a function as /C_STRING, etc. then you can pass sequence values
-- to /c_func or /c_proc and it will automatically poke the values into memory and pass the pointer
-- to the function instead. The opposite also works if the return type of a function is /C_STRING,
-- etc.</dd>
--
--<dd>The C string will be marked for automatic cleanup (see /allocate_string), so if your C function
-- expects to keep the string in memory (which is quite uncommon), you will need to revert to /C_POINTER
-- and allocate the string yourself.</dd>
--</dl>
--
--See also: /define_c_func, /define_c_proc, /c_proc
override function c_func( integer rid, sequence args = {} )
	
	-- get the argument types and return type
	sequence arg_types = map:get( m_args, rid, {} )
	atom return_type = map:get( m_retv, rid, 0 )
	
	-- make sure we have the correct number of arguments
	if length( args ) = length( arg_types ) then
		
		for i = 1 to length( args ) do
			
			-- allocate strings as necessary
			
			if sequence( args[i] ) and arg_types[i] = C_STRING then
				args[i] = allocate_string( args[i], 1 )
			elsif sequence( args[i] ) and arg_types[i] = C_WSTRING then
				args[i] = allocate_wstring( args[i], 1 )
			end if
			
		end for
		
	end if
	
	-- call the function
	object result = eu:c_func( rid, args )
	
	-- fetch strings as necessary
	if return_type = C_STRING then
		result = peek_string( result )
	elsif return_type = C_WSTRING then
		result = peek_wstring( result )
	end if
	
	return result
end function

--/topic Linking to external code
--/proc c_proc( integer rid, sequence args = {} )
--/info
--Call a C function, machine code routine, or translated/compiled Euphoria routine by routine id.
--
--See also: /define_c_func, /define_c_proc, /c_func
override procedure c_proc( integer rid, sequence args = {} )
	-- just call c_func and discard the result
	object result = dll:c_func( rid, args )
end procedure

--/topic Linking to external code
--/type callback( atom x )
--/info
--A callback routine type. Returns zero if /b x is not a valid address returned by /call_back.
--
--See also: /callback_id, /call_back
public type callback( atom x )
	return map:has( m_cbid, x )
end type

--/topic Linking to external code
--/func call_back( object id )
--/info
--Get a machine address for an Euphoria procedure.
--
--<dl>
--<dt> /b Parameters </dt>
--
--<ol>
--<li> /b id : an /b object, either the id returned by /b routine_id, or a pair {'+', id}.</li>
--</ol>
--
--<dt> /b Returns </dt>
--
--<dd>An /b atom, the address of the machine code of the routine. It can be used by Windows, or an
-- external C routine in a Windows .dll or Unix-like shared library (.so) as a 32-bit "call-back"
-- address for calling your Euphoria routine.</dd>
--
--<dt> /b Comments </dt>
--<dd>By default, your routine will work with the stdcall convention. On Windows, you can specify its
-- id as {'+', id}, in which case it will work with the CDECL calling convention instead. On non-Microsoft
-- platforms, you should only use simple IDs, as there is just one standard calling convention, i.e. DECL.</dd>
--
--<dd>You can set up as many call-back functions as you like, but they must all be Euphoria functions
-- (or types) with 0 to 9 arguments. If your routine has nothing to return (it should really be a
-- procedure), just return 0 (say), and the calling C routine can ignore the result.</dd>
--
--<dd>When your routine is called, the argument values will all be 32-bit unsigned (positive) values.
-- You should declare each parameter of your routine as atom, unless you want to impose tighter checking.
-- Your routine must return a 32-bit integer value.</dd>
--
--<dd>You can also use a call-back address to specify a Euphoria routine as an exception handler in
-- the Linux/FreeBSD signal() function. For example, you might want to catch the SIGTERM signal, and
-- do a graceful shutdown. Some Web hosts send a SIGTERM to a CGI process that has used too much CPU
-- time.</dd>
--</dl>
--
--See also: /callback, /callback_id
public function call_back( object id )
	
ifdef WINDOWS and DLL_CDECL then
	-- always use CDECL
	if atom( id ) then
		id = {'+', id}
	end if
end ifdef
	
	atom cb = std_dll:call_back( id )
	map:put( m_cbid, cb, id )
	
	return cb
end function

--/topic Linking to external code
--/func callback_id( atom cb )
--/info
--Returns the routine id of the given callback address.
--
--See also: /callback, /callback_id
public function callback_id( atom cb )
	return map:get( m_cbid, cb, -1 )
end function
